import os
import shutil


def directory_walk(source_d, destination_d):
    '''
    Walk a directory structure and yield full parallel source and destination
    files, munging filenames as necessary
    '''
    for dirpath, dirnames, filenames in os.walk(source_d):
        relpath = os.path.relpath(dirpath, source_d)
        if relpath == '.':
            relpath = ''  # remove implied '.'
        for filename in filenames:
            suffix = filename
            if relpath:
                suffix = os.path.join(relpath, filename)
            full_source_path = os.path.join(source_d, suffix)
            full_destination_path = os.path.join(destination_d, suffix)
            yield full_source_path, full_destination_path


def recursive_symlink_dirs(source_d, destination_d):
    '''
    Create dirs and symlink all files recursively from source_d, ignoring
    errors (e.g. existing files)
    '''
    func = os.symlink
    if os.name == 'nt':
        # NOTE: need to verify that default perms only allow admins to create
        # symlinks on Windows
        func = shutil.copy
    if os.path.exists(destination_d):
        os.rmdir(destination_d)
    shutil.copytree(source_d, destination_d, copy_function=func)


def recursive_hardlink_dirs(source_d, destination_d):
    '''
    Same as above, except creating hardlinks for all files
    '''
    func = os.link
    if os.name == 'nt':
        func = shutil.copy
    if os.path.exists(destination_d):
        os.rmdir(destination_d)
    shutil.copytree(source_d, destination_d, copy_function=func)



from . import views

urls = {
    '': views.conversion_tester_root,
    'conversion/': views.conversion_tester,
    'graph/': views.conversion_graph_root,
    'graph/<ext>/': views.conversion_graph,
    'ajax/workers/': views.poll_worker_queue,
}



import json

import omnic
from omnic.builtin.types.nodejs import NODE_PACKAGE
from omnic.conversion import converter
from omnic.web.viewer import VIEWER_EXT

PACKAGE_JSON_TEMPLATE = {
    'name': 'omnicviewer',
    'version': omnic.__version__,
    'description': 'Auto-generated viewer package',
    'main': 'index.js',
    'dependencies': None,
    'author': '',
    'license': 'GPL-3.0'
}


LOADED_JS = '''
window._OMNIC_VIEWER_BUNDLE_IS_LOADED = true;
'''


def generate_index_js(viewer_names):
    requires = '\n'.join('require("%s");' % name for name in viewer_names)
    return '%s\n%s' % (requires, LOADED_JS)


class ViewerNodePackageBuilder(converter.Converter):
    inputs = [
        VIEWER_EXT.upper(),
    ]
    outputs = [
        str(NODE_PACKAGE),
    ]

    async def convert(self, in_resource, out_resource):

        # Read in dependencies object from the in resource
        with in_resource.cache_open('r') as fd:
            dependencies = json.load(fd)

        # Assemble the package json and write to file
        package_json_dict = dict(PACKAGE_JSON_TEMPLATE)
        package_json_dict['dependencies'] = dependencies
        with out_resource.cache_open_as_dir('package.json', 'w') as fd:
            json.dump(package_json_dict, fd, indent=4)

        # Write the index JavaScript file that serves as the entry point for
        # all viewers
        with out_resource.cache_open_as_dir('index.js', 'w') as fd:
            fd.write(generate_index_js(dependencies.keys()))



from base64 import b64decode

from omnic.responses import placeholder

TRANSPARENT_PNG_PIXEL_B64 = '''
    iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII=
'''
TRANSPARENT_PNG_PIXEL_BYTES = b64decode(TRANSPARENT_PNG_PIXEL_B64.strip())


class PNGPixel(placeholder.BytesPlaceholder):
    types = [
        'PNG',
        'JPG',
        'JPEG',
        'JPE',
        'BMP',
        'GIF',
        'PCX',
        'WEBP',
        'image/png',
        'image/jpeg',
        'image/bmp',
        'image/gif',
        'image/pcx',
        'image/webp',
        'thumb.png',
        'thumb.jpg',
    ]
    content_type = 'image/png'
    bytes = TRANSPARENT_PNG_PIXEL_BYTES


class JSGlobalVar(placeholder.BytesPlaceholder):
    types = [
        'JS',
        'application/javascript',
        'min.js',
        'es5.js',
        'dev-es5.js',
        'bundle.js',
    ]
    content_type = 'application/javascript'
    bytes = b'window._OMNIC_EMPTY = true;\n'


class EmptyAll(placeholder.BytesPlaceholder):
    types = all
    content_type = 'application/x-empty'
    bytes = b''



"""
Tests for `detectors` module.
"""
import os
import tempfile

from omnic.builtin.types.core import ManifestDetector
from omnic.builtin.types.nodejs import NodePackageDetector
from omnic.types.detectors import DetectorManager
from omnic.types.typestring import TypeString

from .testing_utils import Magic


class TestManifestDetector:
    def setup_method(self, method):
        self.fd = tempfile.NamedTemporaryFile(
            suffix='manifest.json', delete=False)
        self.d = ManifestDetector()

    def teardown_method(self, method):
        if not self.fd.closed:
            self.fd.close()
        os.remove(self.fd.name)

    def test_can_improve(self):
        assert not self.d.can_improve(TypeString('PNG'))
        assert not self.d.can_improve(TypeString('manifest.json'))
        assert self.d.can_improve(TypeString('application/json'))
        assert self.d.can_improve(TypeString('JSON'))

    def test_can_detect(self):
        assert self.d.can_detect('/path/to/manifest.json')
        assert not self.d.can_detect('/path/to/manifest.json.js')
        assert not self.d.can_detect('/path/to/other.json')

    def test_detect_empty(self):
        assert self.d.detect(self.fd.name) == None

    def test_detect_json(self):
        self.fd.write(b'{"JSON": true}')
        self.fd.close()  # Empty file
        assert str(self.d.detect(self.fd.name)) == 'manifest.json'


class TestNodePackageDetector:
    def setup_method(self, method):
        self.path = tempfile.mkdtemp(prefix='omnic_test_')
        self.pjson_path = os.path.join(self.path, 'package.json')
        self.d = NodePackageDetector()
        self.fd = None

    def teardown_method(self, method):
        if self.fd and not self.fd.closed:
            self.fd.close()
        try:
            os.remove(self.pjson_path)
        except OSError:
            pass
        os.removedirs(self.path)

    def _write_pjson(self):
        self.fd = open(self.pjson_path, 'w+')
        self.fd.write('{}')
        self.fd.close()

    def test_can_improve(self):
        assert not self.d.can_improve(TypeString('PNG'))
        assert self.d.can_improve(TypeString('directory'))

    def test_can_detect(self):
        assert not self.d.can_detect(self.path)
        self._write_pjson()
        assert self.d.can_detect(self.path)

    def test_detect_empty(self):
        assert self.d.detect(self.path) == None

    def test_detect_json(self):
        self._write_pjson()
        assert str(self.d.detect(self.path)) == 'nodepackage'

    def test_detect_node_modules(self):
        self._write_pjson()
        _nm_path = os.path.join(self.path, 'node_modules')
        os.mkdir(_nm_path)
        assert str(self.d.detect(self.path)) == 'installed_nodepackage'
        os.rmdir(_nm_path)  # clean up


class TestBuiltinDetectorsForImages:
    '''
    Integration test for mimetype + extension detectors, using image examples
    '''

    def setup_method(self, method):
        self.fd = tempfile.NamedTemporaryFile(suffix='test.png', delete=False)
        self.detect = DetectorManager().detect

    def teardown_method(self, method):
        os.remove(self.fd.name)

    def test_extension_detector(self):
        self.fd.close()  # Empty file, no possible guess of mimetype, use ext
        ts = self.detect(self.fd.name)
        assert ts.mimetype == 'image/png'

    def test_magic_bytes_detector(self):
        # Even though ends with png, we write JPG magic bytes into it
        self.fd.write(bytes(Magic.JPEG))
        self.fd.close()
        ts = self.detect(self.fd.name)
        assert ts.mimetype == 'image/jpeg'


class TestBuiltinDetectorsForManifest:
    '''
    Integration test for manifest detector
    '''

    def setup_method(self, method):
        self.fd = tempfile.NamedTemporaryFile(
            suffix='manifest.json', delete=False)
        self.detect = DetectorManager().detect

    def teardown_method(self, method):
        os.remove(self.fd.name)

    def test_detects_json(self):
        self.fd.write(b'{"JSON": true}')
        self.fd.close()
        ts = self.detect(self.fd.name)
        assert str(ts) == 'manifest.json'



import hashlib
import os
import shutil
from urllib.parse import urlparse

from omnic import singletons
from omnic.utils.iters import group_by

try:
    import requests
except ImportError:
    requests = None


class Resource:
    '''
    Abstract base class for Resources
    '''
    __slots__ = (
        # Slots used by all Resources
        'url', 'url_string', 'url_path_split', 'url_path_basename',
        'basename', 'md5', 'cache_path_base', 'cache_path'

        # Optional slots used by subclasses
        'typestring', 'foreign', 'data',
    )

    def __init__(self, url, is_url=True):
        # Setup props
        self.url = url

        # Parse and process URL
        self.url_string = url
        self.url = urlparse(url)
        self.url_path_split = self.url.path.split('/')
        self.url_path_basename = self.url_path_split[-1]
        if len(self.url_path_basename) < 1:
            # Path is too small, probably ends with /, try 1 up
            self.url_path_basename = self.url_path_split[-2]
        self.basename = self._get_basename()

        # Generate MD5
        self.md5 = hashlib.md5(self.url_string.encode('utf-8')).hexdigest()

        # Generate filepath
        self.cache_path_base = os.path.join(
            singletons.settings.PATH_PREFIX,
            singletons.settings.RESOURCE_CACHE_INTERFIX,
            *self.path_grouping(),
        )

        self.cache_path = os.path.join(
            self.cache_path_base,
            self.basename,
        )

    def _get_basename(self):
        raise NotImplementedError()

    def path_grouping(self):
        if singletons.settings.PATH_GROUPING is None:
            return ['']
        if singletons.settings.PATH_GROUPING == 'MD5':
            return group_by(self.md5, 8)
        raise singletons.settings.Error('Invalid PATH_GROUPING')

    def cache_makedirs(self, subdir=None):
        '''
        Make necessary directories to hold cache value
        '''
        if subdir is not None:
            dirname = self.cache_path
            if subdir:
                dirname = os.path.join(dirname, subdir)
        else:
            dirname = os.path.dirname(self.cache_path)
        os.makedirs(dirname, exist_ok=True)

    def cache_open(self, mode='rb'):
        if 'w' in mode:
            self.cache_makedirs()
        return open(self.cache_path, mode=mode)

    def cache_open_as_dir(self, filepath, mode='rb'):
        if 'w' in mode:
            self.cache_makedirs(subdir=os.path.dirname(filepath))
        path = os.path.join(self.cache_path, filepath)
        return open(path, mode=mode)

    def cache_exists(self):
        return os.path.exists(self.cache_path)

    def cache_remove(self):
        return os.unlink(self.cache_path)

    def cache_remove_as_dir(self):
        return shutil.rmtree(self.cache_path)

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, repr(self.url_string))

    def __hash__(self):
        return hash(self.url_string)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented


class ForeignResource(Resource):
    '''
    A resource from a foreign source (e.g. a URL), which does not have a known
    type, and may or may not be downloaded.
    '''

    def _get_basename(self):
        return self.url_path_basename

    def download(self):
        if requests is None:
            raise RuntimeError('requests is not installed')
        req = requests.get(self.url_string, stream=True)
        with self.cache_open('wb') as f:
            shutil.copyfileobj(req.raw, f)

    def validate(self):
        if not check_url(singletons.settings, self.url):
            raise URLError('Invalid URL: "%s"' % self.url_string)

    def guess_typed(self):
        if not self.cache_exists():
            raise CacheError('Cannot guess type without first downloaded')
        ts = singletons.detectors.detect(self.cache_path)
        return TypedForeignResource(self.url_string, ts)

    def cache_remove_all(self):
        return shutil.rmtree(self.cache_path_base)


class TypedForeignResource(Resource):
    '''
    A resource freshly downloaded from a foreign source that has a known
    (guessed) type.
    '''

    def __init__(self, url, typestring):
        self.typestring = typestring
        super().__init__(url)

    def __repr__(self):
        return '%s%s' % (
            type(self).__name__,
            repr((self.url_string, self.typestring))
        )

    def _get_basename(self):
        # Ignores URL basename
        base, _ = os.path.splitext(self.url_path_basename)
        return self.typestring.modify_basename(base)

    def symlink_from(self, foreign_resource):
        if foreign_resource.cache_path == self.cache_path:
            return
        os.symlink(foreign_resource.cache_path, self.cache_path)


class TypedResource(Resource):
    '''
    A resource that is or will be the result of a conversion, thus having
    a known type.
    '''

    def __init__(self, url_string, typestring):
        self.typestring = typestring
        super().__init__(url_string)

    def __repr__(self):
        name = type(self).__name__
        tup = (name, repr(self.url_string), repr(self.typestring))
        return '%s(%s, %s)' % tup

    def _get_basename(self):
        return self.typestring.modify_basename(self.url_path_basename)

    def __hash__(self):
        return hash('%s - %s' % (self.url_string, str(self.typestring)))


class TypedLocalResource(Resource):
    '''
    A resource which is treated like a foreign resource, but originates from a
    path instead of an URL.

    Used in CLI conversions.
    '''

    def __init__(self, path, typestring=None):
        self.path = path

        if typestring:
            # TypeString specified: not foreign, should munge path
            self.foreign = False
        else:
            # No TypeString specified: foreign, keep path but guess type
            self.foreign = True
            typestring = singletons.detectors.detect(path)

        self.typestring = typestring
        super().__init__('file://%s' % path)
        if self.foreign:
            self.cache_path = self.path

    def _get_basename(self):
        if self.foreign:
            return os.path.basename(self.path)
        else:
            return self.typestring.modify_basename(self.url_path_basename)

    def __hash__(self):
        return hash(self.path)


class TypedPathedLocalResource(TypedLocalResource):
    '''
    Like a TypedLocalResource, but is also opinionated about where it should be
    stored, typically somewhere other than the cache.

    Used in CLI conversions.
    '''

    def __init__(self, path, typestring):
        super().__init__(path, typestring)
        self.cache_path = os.path.join(
            os.path.dirname(path),
            self._get_basename(),
        )

    def _get_basename(self):
        base, _ = os.path.splitext(self.path)
        return self.typestring.modify_basename(base)


class ForeignBytesResource(ForeignResource):
    '''
    A foreign resource that consists of a string of bytes instead of a URL.
    '''

    def __init__(self, data, extension=None, basename='source'):
        self.data = data
        ext = '.%s' % extension if extension else ''
        md5 = hashlib.md5(data).hexdigest()
        virtual_path = 'file://%s/%s%s' % (md5, basename, ext)
        super().__init__(virtual_path)

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, repr(self.data))

    def save(self):
        with self.cache_open('wb') as f:
            f.write(self.data)

    def download(self):
        self.save()

    def validate(self):
        # For now, validate all foreign string resources (should eventually add
        # security measures here)
        pass


class URLError(ValueError):
    pass


class CacheError(RuntimeError):
    pass


def check_url(config, url):
    if config.ALLOWED_LOCATIONS == '*':
        return True
    return url.netloc in config.ALLOWED_LOCATIONS



"""
Tests for `typestring` module.
"""


from omnic.types.typestring import TypeString


class TestMimeTypeTypeString:
    @classmethod
    def setup_class(cls):
        cls.ts = TypeString('application/json')

    def test_is_mimetype_method(self):
        assert self.ts.mimetype == 'application/json'
        assert self.ts.extension == 'JSON'
        assert not self.ts.is_qualifier

    def test_str(self):
        assert str(self.ts) == 'application/json'

    def test_repr(self):
        assert repr(self.ts) == 'TypeString(\'application/json\')'

    def test_equality(self):
        assert self.ts == TypeString('application/json')

    def test_has_no_arguments(self):
        assert self.ts.arguments == tuple()

    def test_modify_basename(self):
        assert self.ts.modify_basename('thing.xml') == 'thing.xml.json'
        assert self.ts.modify_basename('thing') == 'thing.json'


class TestExtensionTypeString:
    @classmethod
    def setup_class(cls):
        cls.ts = TypeString('JPEG')

    def test_is_mimetype_method(self):
        assert self.ts.mimetype == 'image/jpeg'
        assert self.ts.extension == 'JPEG'
        assert not self.ts.is_qualifier

    def test_equality(self):
        assert self.ts == TypeString('JPEG')
        assert self.ts != TypeString('jpeg')

    def test_str(self):
        assert str(self.ts) == 'JPEG'

    def test_has_no_arguments(self):
        assert self.ts.arguments == tuple()

    def test_modify_basename(self):
        assert self.ts.modify_basename('thing.xml') == 'thing.xml.jpeg'
        assert self.ts.modify_basename('thing') == 'thing.jpeg'


class TestQualifierArgumentsTypeString:
    @classmethod
    def setup_class(cls):
        cls.ts = TypeString('thumb.png:400x300')

    def test_is_mimetype_method(self):
        assert self.ts.mimetype is None
        assert self.ts.extension is None
        assert self.ts.is_qualifier

    def test_equality(self):
        assert self.ts == TypeString('thumb.png:400x300')
        assert self.ts != TypeString('thumb.png:300x300')

    def test_str(self):
        assert str(self.ts) == 'thumb.png:400x300'

    def test_has_arguments(self):
        assert self.ts.arguments == ('400x300',)

    def test_modify_basename(self):
        assert self.ts.modify_basename(
            'thing.xml') == 'thing.xml.400x300.thumb.png'
        assert self.ts.modify_basename('thing') == 'thing.400x300.thumb.png'



from omnic.conversion import converter


class MeshLabConverter(converter.ExecConverter):
    inputs = [
        # All 3D model formats and extensions supported by meshlabserver
        'OBJ',
        'MESH',
        'X3D',
        'X3DV',
        'X3DB',
        'OFF',
        '3DS',
        'PTX',
        'PLY',
        'DAE',
        'XYZ',
        'APTS',
        'PTS',
        'TRI',
        'ASC',
        'VRML',
        'ALN',
        'model/mesh',
        'model/vnd',
        'model/x3d',
        'model/vrml',
        'model/x3d+vrml',
        'model/x3d+xml',
        'model/x3d+binary',
    ]

    outputs = [
        # Just convert everything into basic STL's for now
        'STL',
    ]

    command = [
        'meshlabserver',
        '-i',
        '$IN',
        '-o',
        '$OUT',
    ]


class Jsc3dRenderer(converter.ExecConverter):
    inputs = [
        'STL',
        'OBJ',
    ]

    outputs = [
        'PNG',
    ]

    command = [
        'jsc3d',
        '$IN',
        '$OUT',
    ]


class DimeConverter(converter.ExecConverter):
    # NOTE: Untested
    # Uses DIME converter
    inputs = [
        'DXF',
        'image/vnd.dxf',
    ]

    outputs = [
        'VRML',
        'model/vrml',
    ]

    command = [
        'dxf2vrml',
        '$IN',
        '-o',
        '$OUT',
    ]


class KabejaAutocadConverter(converter.ExecConverter):
    # NOTE: Untested
    # NOTE: Kabeja dies if there are spaces in the file name (yeap.... >_>),
    # need to add feature to converter base class to only have pretty/clean
    # filenames
    inputs = [
        'DXF',
        'image/vnd.dxf',
    ]

    outputs = [
        'SVG',
    ]

    command = [
        'kabeja',
        '-nogui',
        '-pipeline',
        'svg',
        '$IN',
        '$OUT',
    ]



"""
Tests for `converter` module.
"""

import os
import tempfile
from unittest.mock import MagicMock, patch

import pytest

from omnic import singletons
from omnic.config.utils import use_settings
from omnic.conversion import converter
from omnic.conversion.graph import ConverterGraph
from omnic.types.resource import TypedResource
from omnic.types.typestring import TypeString
from omnic.utils.graph import NoPath

from .testing_utils import (AgreeableDetector, DummyDetector, Magic,
                            rm_tmp_files)

URL = 'http://mocksite.local/file.png'
DIR_URL = 'http://mocksite.local/my_files'

# TODO: Fix these tests to be less integrate-y, mock out subprocess
# calls


class HardLinkConverter(converter.HardLinkConverter):
    inputs = ['JPEG']
    outputs = ['JPG']


class ExecConverter(converter.ExecConverter):
    inputs = ['JPEG']
    outputs = ['JPG']
    command = [
        'mv',
        '$IN',
        '$OUT',
    ]


class ExecConverterWithArgs(ExecConverter):
    command = [
        'cp',
        '$0',
        '$IN',
        '$OUT',
    ]


class ExecConverterWithOutputFilename(ExecConverter):
    command = [
        'mv',
        '$IN',
        '_tmp_output_file',
    ]

    def get_output_filename(self, in_res, out_res):
        return '_tmp_output_file'


class UnavailableConverter(converter.Converter):
    inputs = ['in2']
    outputs = ['out2']

    @staticmethod
    def configure():
        raise converter.ConverterUnavailable()


class AvailableConverter(converter.Converter):
    inputs = ['in']
    outputs = ['out']


class MockNpmConverter(converter.AdditiveDirectoryExecConverter):
    inputs = ['nodepackage']
    outputs = ['installed_nodepackage']
    command = ['mkdir', 'node_modules']  # Mock npm install


# Set up system of dummy converters
class ConvertMovieToImage(converter.HardLinkConverter):
    inputs = ['MOV', 'AVI', 'MP4']
    outputs = ['JPG']


class ConvertImageToThumb(converter.HardLinkConverter):
    inputs = ['JPG', 'PNG', 'GIF']
    outputs = ['thumb.png']


class Convert3DGraphicsToMovie(converter.HardLinkConverter):
    inputs = ['STL', 'OBJ', 'MESH']
    outputs = ['AVI']


class Convert3DGraphicsToImage(converter.HardLinkConverter):
    inputs = ['STL', 'OBJ', 'MESH']
    outputs = ['PNG']


class CleanUpAudio(converter.HardLinkConverter):
    inputs = ['MP3', 'WAV', 'OGG']
    outputs = ['cleaned.ogg']


class MockConfig:
    PATH_GROUPING = 'MD5'
    PATH_PREFIX = ''
    ALLOWED_LOCATIONS = '*'
    CONVERTERS = [
        ConvertMovieToImage,
        ConvertImageToThumb,
        Convert3DGraphicsToMovie,
        Convert3DGraphicsToImage,
        CleanUpAudio,
    ]


class ConverterTestBase:
    def setup_method(self, method):
        MockConfig.PATH_PREFIX = tempfile.mkdtemp()
        singletons.settings.use_settings(MockConfig)
        self.res = TypedResource(URL, TypeString('JPEG'))
        self.res2 = TypedResource(URL, TypeString('JPG'))
        with self.res.cache_open('wb') as f:
            f.write(Magic.JPEG)

    def teardown_method(self, method):
        singletons.settings.use_previous_settings()
        rm_tmp_files(
            self.res.cache_path,
            self.res2.cache_path,
        )

    def _path(self, in_str, out_str):
        return self.cgraph.find_path(TypeString(in_str), TypeString(out_str))

    def _check_convert(self):
        self.converter.convert_sync(self.res, self.res2)
        assert self.res2.cache_exists()
        with self.res2.cache_open() as f:
            assert f.read() == Magic.JPEG


class DirectoryConverterTestBase:
    def setup_method(self, method):
        MockConfig.PATH_PREFIX = tempfile.mkdtemp()
        singletons.settings.use_settings(MockConfig)
        self.res = TypedResource(DIR_URL, TypeString('nodepackage'))
        self.res2 = TypedResource(DIR_URL, TypeString('installed_nodepackage'))
        with self.res.cache_open_as_dir('package.json', 'wb') as f:
            f.write(b'{}\n')
        with self.res.cache_open_as_dir('lib/main.js', 'wb') as f:
            f.write(b'console.log("test stuff");\n')

    def teardown_method(self, method):
        singletons.settings.use_previous_settings()
        rm_tmp_files(
            'package.json',
            'lib/main.js',
            prefixes=[self.res.cache_path, self.res2.cache_path],
        )


class TestAdditiveDirectoryExecConverter(DirectoryConverterTestBase):
    def test_convert(self):
        self.converter = MockNpmConverter()
        self.converter.convert_sync(self.res, self.res2)
        j = os.path.join
        assert self.res2.cache_exists()

        # ensure only new directory has the 'node_modules'
        assert os.path.isdir(j(self.res2.cache_path, 'node_modules'))
        assert not os.path.isdir(j(self.res.cache_path, 'node_modules'))
        assert not os.path.islink(j(self.res2.cache_path, 'lib'))
        assert not os.path.islink(j(self.res2.cache_path, 'node_modules'))

        # Ensure its just hardlinked and not copied
        assert not os.path.islink(j(self.res2.cache_path, 'package.json'))
        assert os.stat(j(self.res2.cache_path, 'package.json')).st_nlink == 2

        with self.res2.cache_open_as_dir('package.json') as f:
            assert f.read() == b'{}\n'
        with self.res2.cache_open_as_dir('lib/main.js') as f:
            assert f.read() == b'console.log("test stuff");\n'


class TestHardLinkConverter(ConverterTestBase):
    def test_convert(self):
        self.converter = HardLinkConverter()
        self._check_convert()


class TestDetectorConverter(ConverterTestBase):
    def test_convert_nothing(self):
        self.converter = converter.DetectorConverter()
        self.converter.detector = AgreeableDetector
        with pytest.raises(converter.ConversionInputError):
            # Wrong direction (from nonexistent to existent)
            self.converter.convert_sync(self.res2, self.res)

    def test_convert_invalid(self):
        self.converter = converter.DetectorConverter()
        self.converter.detector = DummyDetector
        with pytest.raises(converter.ConversionInputError):
            self.converter.convert_sync(self.res, self.res2)

    def test_convert_success(self):
        self.converter = converter.DetectorConverter()
        self.converter.detector = AgreeableDetector
        self._check_convert()  # valid conversion


class TestDirectoryConverter(DirectoryConverterTestBase):
    pass


class TestExecConverter(ConverterTestBase):
    def test_convert(self):
        self.converter = ExecConverter()
        self._check_convert()

    def test_convert_with_arg(self):
        # cp -s creates a symbolic link
        self.res2 = TypedResource(URL, TypeString('JPG:-s'))
        self.converter = ExecConverterWithArgs()
        self._check_convert()
        assert os.path.islink(self.res2.cache_path)

    def test_convert_with_custom_filename(self):
        self.res2 = TypedResource(URL, TypeString('JPG'))
        self.converter = ExecConverterWithOutputFilename()
        self._check_convert()


class TestExecConverterUnitTest:
    class Subclass(converter.ExecConverter):
        command = ['test', '$IN', '$OUT']

    def _get_mocked_resource(self, ts):
        res = TypedResource(URL, TypeString(ts))
        res.cache_makedirs = MagicMock()
        res.cache_path = 'test/path.%s' % ts
        return res

    def test_configure_failure(self):
        shutil = {'which.return_value': None}
        with patch('omnic.conversion.converter.shutil', **shutil):
            with pytest.raises(converter.ConverterUnavailable):
                self.Subclass.configure()

    def test_configure_succeed(self):
        shutil = {'which.return_value': '/bin/test'}
        with patch('omnic.conversion.converter.shutil', **shutil):
            self.Subclass.configure()

    def test_convert_sync(self):
        res1 = self._get_mocked_resource('input')
        res2 = self._get_mocked_resource('output')
        sb = self.Subclass()
        info = {'run.return_value': '/bin/test'}
        with patch('omnic.conversion.converter.subprocess', **info) as sp:
            sb.convert_sync(res1, res2)
        res1.cache_makedirs.assert_called_once_with()
        res2.cache_makedirs.assert_called_once_with()
        sp.run.assert_called_once_with(
            ['test', 'test/path.input', 'test/path.output'],
            cwd='test',
        )


class TestBasicConverterGraph(ConverterTestBase):
    @classmethod
    def setup_class(cls):
        cls.cgraph = ConverterGraph(MockConfig.CONVERTERS)

    def test_find_path(self):
        # Test typical case
        results = self._path('AVI', 'thumb.png:200x200')
        assert len(results) == 2  # should be 2 steps
        assert all(len(step) == 3 for step in results)  # each step should be 3
        assert all(
            isinstance(step[1], TypeString) and isinstance(step[2], TypeString)
            for step in results  # Ensure we are getting typestrings
        )
        assert results[0][0] is ConvertMovieToImage
        assert results[1][0] is ConvertImageToThumb

    def test_finds_shortest_path(self):
        # Ensure not taking long route
        results = self._path('STL', 'thumb.png:100x100')
        assert len(results) == 2  # should be 2 steps
        assert all(len(step) == 3 for step in results)  # each step should be 3
        assert results[0][0] is Convert3DGraphicsToImage
        assert results[1][0] is ConvertImageToThumb


class TestConverterGraphCustomPaths:
    def teardown_method(self, method):
        singletons.settings.use_previous_settings()

    def _path(self, in_str, out_str):
        return self.cgraph.find_path(TypeString(in_str), TypeString(out_str))

    def test_pruned_converters(self):
        cgraph = ConverterGraph([
            AvailableConverter,
            UnavailableConverter,
        ], prune_converters=True)
        assert len(cgraph.converters) == 1

    @use_settings(preferred_conversion_paths=[('STL', 'AVI', 'JPG', 'thumb.png')])
    def test_preferred_conversions(self):
        self.cgraph = ConverterGraph(MockConfig.CONVERTERS)
        results = self._path('STL', 'thumb.png')
        assert len(results) == 3  # Forcing it to 1 extra step

    @use_settings(conversion_profiles={'thumb': 'thumb.png:123x456'})
    def test_simple_conversion_profile(self):
        self.cgraph = ConverterGraph(MockConfig.CONVERTERS)
        results = self._path('STL', 'thumb')
        assert len(results) == 2  # should be 2 steps
        assert all(len(step) == 3 for step in results)  # each step should be 3
        assert results[0][0] is Convert3DGraphicsToImage
        assert results[1][0] is ConvertImageToThumb
        assert results[1][2].arguments == ('123x456', )

    @use_settings(conversion_profiles={
        'avithumb': ('AVI', 'JPG', 'thumb.png')})
    def test_conversion_profile_path(self):
        self.cgraph = ConverterGraph(MockConfig.CONVERTERS)
        results = self._path('STL', 'avithumb')
        assert len(results) == 3  # Ensure it is 1 "extra" step
        assert all(len(step) == 3 for step in results)  # each step should be 3
        assert results[0][0] is Convert3DGraphicsToMovie
        assert results[1][0] is ConvertMovieToImage
        assert results[2][0] is ConvertImageToThumb

    @use_settings(conversion_profiles={'avithumb': ('AVI', 'OBJ')})
    def test_invalid_conversion_profile(self):
        self.cgraph = ConverterGraph(MockConfig.CONVERTERS)
        assert not self.cgraph.conversion_profiles.keys()
        with pytest.raises(NoPath):
            self._path('STL', 'avithumb')

    @use_settings(conversion_profiles={
        'avithumb': ('AVI', 'JPG:frame=2', 'thumb.png:123x456')})
    def test_conversion_profile_path_with_internal_args(self):
        self.cgraph = ConverterGraph(MockConfig.CONVERTERS)
        results = self._path('STL', 'avithumb')
        assert len(results) == 3  # Ensure it is 1 "extra" step
        assert all(len(step) == 3 for step in results)  # each step should be 3
        assert results[0][0] is Convert3DGraphicsToMovie
        assert results[1][0] is ConvertMovieToImage
        assert results[2][0] is ConvertImageToThumb
        assert results[1][2].arguments == ('frame=2', )
        assert results[2][2].arguments == ('123x456', )



__author__ = 'michaelb'
__email__ = 'michaelpb@gmail.com'
__version__ = '0.1.12'

# flake8: noqa

# Set up singleton system
from omnic.utils.singleton import SingletonManager
singletons = SingletonManager()

# Ensure settings gets registered as a singleton
from omnic.config.settingsmanager import SettingsManager
singletons.register('settings', SettingsManager)



from omnic import singletons
from omnic.types.resource import (ForeignResource, TypedForeignResource,
                                  TypedResource)
from omnic.types.typestring import TypeString


async def multiconvert(url_string, to_type, enqueue_convert):
    '''
    Given a URL string that has already been downloaded, enqueue
    necessary conversion to get to target type
    '''
    target_ts = TypeString(to_type)
    foreign_res = ForeignResource(url_string)

    # Determine the file type of the foreign resource
    typed_foreign_res = foreign_res.guess_typed()

    if not typed_foreign_res.cache_exists():
        # Symlink to new location that includes typed extension
        typed_foreign_res.symlink_from(foreign_res)

    # Now find path between types
    original_ts = typed_foreign_res.typestring
    path = singletons.converter_graph.find_path(original_ts, target_ts)

    # Loop through each step in graph path and convert
    is_first = True
    for converter_class, from_ts, to_ts in path:
        converter = converter_class()
        in_resource = TypedResource(url_string, from_ts)
        if is_first:  # Ensure first resource is just the source one
            in_resource = TypedForeignResource(url_string, from_ts)
        out_resource = TypedResource(url_string, to_ts)
        await enqueue_convert(converter, in_resource, out_resource)
        is_first = False



import asyncio
import json

import aiohttp

from omnic import singletons
from omnic.builtin.types.core import DIRECTORY, MANIFEST_JSON
from omnic.conversion import converter


class ManifestDownloader(converter.Converter):
    # TODO: Possibly switch to using binary system packages like curl?
    inputs = [
        str(MANIFEST_JSON),
    ]
    outputs = [
        str(DIRECTORY),
    ]

    async def convert(self, in_resource, out_resource):
        # TODO: Create Manifest JSON helper util, support TXT format too
        with in_resource.cache_open('r') as fd:
            parsed = json.load(fd)
        files = parsed['files'].items()

        # Download all files with the same client session
        self._new_aiohttp_client()
        self.file_descriptors = {
            relpath: out_resource.cache_open_as_dir(relpath, 'wb')
            for relpath, file_url in files
        }
        await asyncio.wait([
            self._download_async(file_url, self.file_descriptors[relpath])
            for relpath, file_url in files
        ])
        self._close()

    # TODO: move all this to a helper class
    def _new_aiohttp_client(self):
        loop = singletons.eventloop.loop
        self.aiohttp = aiohttp.ClientSession(loop=loop)

    def _close(self):
        '''
        Closes aiohttp session and all open file descriptors
        '''
        if hasattr(self, 'aiohttp'):
            if not self.aiohttp.closed:
                self.aiohttp.close()
        if hasattr(self, 'file_descriptors'):
            for fd in self.file_descriptors.values():
                if not fd.closed:
                    fd.close()

    async def _download_async(self, url, f_handle):
        DOWNLOAD_TIMEOUT = 10
        DOWNLOAD_CHUNK_SIZE = 1024
        with aiohttp.Timeout(DOWNLOAD_TIMEOUT):
            async with self.aiohttp.get(url) as response:
                while True:
                    chunk = await response.content.read(DOWNLOAD_CHUNK_SIZE)
                    if not chunk:
                        break
                    f_handle.write(chunk)
                return await response.release()



"""
Tests for `config` module.
"""
import os
import tempfile

import pytest

from omnic.config.exceptions import ConfigurationError
from omnic.config.settingsmanager import SettingsManager
from omnic.config.utils import use_settings

TEST_SETTING = 123


class ExampleClassA:
    pass


class ExampleClassB:
    pass


class TestSettings:
    def test_overriding(self):
        settings = SettingsManager()
        assert settings.SERVICES

        class MockSettings:
            SERVICES = []
        settings.use_settings(MockSettings)
        assert not settings.SERVICES
        settings.use_previous_settings()
        assert settings.SERVICES

    def test_overriding_with_dict(self):
        settings = SettingsManager()
        assert settings.SERVICES
        settings.set(services=[])
        assert not settings.SERVICES
        settings.use_settings_dict({'services': ['asdf']})
        assert not settings.SERVICES
        settings.use_previous_settings()
        assert not settings.SERVICES

    def test_overriding_exception(self):
        settings = SettingsManager()
        with pytest.raises(ValueError):
            settings.set(SERVICES=[])
        with pytest.raises(AttributeError):
            settings.set(nonexistant=[])

    def test_set(self):
        settings = SettingsManager()
        assert settings.SERVICES

        class MockSettings:
            SERVICES = []
        settings.use_settings(MockSettings)
        assert not settings.SERVICES
        settings.use_previous_settings()
        assert settings.SERVICES

    def test_catching_lowercase_errors(self):
        settings = SettingsManager()

        class MockSettings:
            private_thing = 123
        settings.use_settings(MockSettings)
        with pytest.raises(AttributeError):
            settings.private_thing


class TestSettingsCustom:
    def setup_method(self, method):
        self._original = os.environ.get('OMNIC_SETTINGS', None)
        self._original_cwd = os.getcwd()
        if self._original:
            del os.environ['OMNIC_SETTINGS']

    def teardown_method(self, method):
        if not self._original:
            if 'OMNIC_SETTINGS' in os.environ:
                del os.environ['OMNIC_SETTINGS']
        else:
            os.environ['OMNIC_SETTINGS'] = self._original
        os.chdir(self._original_cwd)

    def test_custom_with_environ(self):
        # Simply use self as a test settings
        os.environ['OMNIC_SETTINGS'] = 'test.test_config'
        settings = SettingsManager()
        assert settings.TEST_SETTING == 123

    def test_custom_with_settings_py(self):
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            with open(os.path.join(tempdir, 'settings.py'), 'w+') as f:
                f.write('SOME_TEST_VALUE = 1337')
            settings = SettingsManager()
            assert settings.SOME_TEST_VALUE == 1337


class TestSettingsLoading:
    def test_loading_modules(self):
        # Simply use self as a test settings
        settings = SettingsManager()

        class MockSettings:
            SERVICES = [
                'test.test_config',
            ]
        settings.use_settings(MockSettings)
        assert len(settings.load_all('SERVICES')) == 1
        assert hasattr(settings.load_all('SERVICES')[0], 'TEST_SETTING')

    def test_loading_class_modules(self):
        # Simply use self as a test settings
        settings = SettingsManager()

        class MockSettings:
            SERVICES = [
                'test.test_config.ExampleClassA',
            ]
        settings.use_settings(MockSettings)
        assert len(settings.load_all('SERVICES')) == 1
        assert settings.load_all('SERVICES')[0] is ExampleClassA

    def test_loading_exceptions(self):
        # Simply use self as a test settings
        settings = SettingsManager()

        class MockSettings:
            SERVICES = [
                'test.test_config.DoesntExist',
            ]
        settings.use_settings(MockSettings)
        with pytest.raises(ConfigurationError):
            settings.load_all('SERVICES')

    def test_loading_modules_dict(self):
        # Simply use self as a test settings
        settings = SettingsManager()

        class MockSettings:
            SERVICES = {
                'a': 'test.test_config.ExampleClassA',
                'b': 'test.test_config.ExampleClassB',
            }
        settings.use_settings(MockSettings)
        assert len(settings.load_all('SERVICES')) == 2
        assert settings.load_all('SERVICES')['a'] is ExampleClassA
        assert settings.load_all('SERVICES')['b'] is ExampleClassB

    def test_loading_with_default(self):
        # Simply use self as a test settings
        settings = SettingsManager()

        class MockSettings:
            SERVICES = [
                'test.test_config.ExampleClassA',
                'test.test_config.DoesntExist',
            ]

        class MockLoader(str):
            pass

        settings.use_settings(MockSettings)
        results = settings.load_all('SERVICES', default=MockLoader)
        assert len(results) == 2
        assert results[0] is ExampleClassA
        assert isinstance(results[1], MockLoader)
        assert str(results[1]) == str(MockSettings.SERVICES[1])

    def test_path_search(self):
        settings = SettingsManager()
        results = settings.import_path_to_absolute_path(
            'omnic.config',
            '__init__.py',
        )
        assert results == os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            'omnic',
            'config',
        )

    def test_path_search_exception(self):
        settings = SettingsManager()
        with pytest.raises(ConfigurationError):
            settings.import_path_to_absolute_path(
                'omnic.config',
                'doesnotexist.txt',
            )


class TestSettingsUtils:
    @use_settings(services=[])
    def test_overriding_with_decorator(self):
        from omnic import singletons
        assert not singletons.settings.SERVICES

    def test_overriding_with_context_manager(self):
        from omnic import singletons
        settings = singletons.settings
        assert settings.SERVICES
        with use_settings(services=[]):
            assert not settings.SERVICES
        assert settings.SERVICES



from omnic.conversion import converter


class InkscapeConverter(converter.ExecConverter):
    inputs = [
        'SVGZ',
        'DIA',
        'AI',
        'WPF',
        'WPM',
        'SK1',
        'PLT',
        'OUTLINE',

        'image/svg+xml'
    ]

    outputs = [
        'SVG',
    ]

    command = [
        'inkscape',
        '-z',
        '-f',
        '$IN',
        '-l',  # means: Export plain SVG
        '$OUT',
    ]


class InkscapeRasterizer(converter.ExecConverter):
    inputs = [
        'SVG',
        'image/svg+xml'
    ]

    outputs = [
        'PNG',
    ]

    command = [
        'inkscape',
        '-z',
        '-f',
        '$IN',
        '-e',  # means: Export PNG
        '$OUT',
    ]



from omnic import singletons


class PlaceholderNotFound(ValueError):
    pass


class Placeholder:
    types = []
    mimetype = 'application/x-empty'

    def __init__(self, typestring):
        self.typestring = typestring

    @classmethod
    def matches(cls, typestring):
        if cls.types is all:
            return True
        return str(typestring.ts_format) in cls.types

    async def stream_response(self, response):
        pass


class BytesPlaceholder(Placeholder):
    bytes = bytes([0])

    async def stream_response(self, response):
        response.write(self.bytes)


class PlaceholderSelector:
    '''
    Helper singleton that takes in all placeholders
    '''

    def __init__(self, use_settings=None):
        if use_settings:
            # Used by testing, TODO remove
            self.placeholders = use_settings.PLACEHOLDERS
        else:
            self.placeholders = singletons.settings.load_all('PLACEHOLDERS')

    def get_placeholder(self, typestring):
        for placeholder_class in self.placeholders:
            if placeholder_class.matches(typestring):
                return placeholder_class(typestring)
        return None

    def stream_response(self, typestring, response):
        placeholder = self.get_placeholder(typestring)
        if not placeholder:
            raise PlaceholderNotFound(str(typestring))
        return response.stream(
            placeholder.stream_response,
            content_type=placeholder.content_type,
        )


singletons.register('placeholders', PlaceholderSelector)



import random

from jinja2 import Environment, PackageLoader, select_autoescape

from omnic import singletons


class Jinja2TemplateHelper:
    def __init__(self, package, path):
        self.env = Environment(
            loader=PackageLoader(package, path),
            autoescape=select_autoescape(['html'])
        )

    def process_context(self, request):
        return {
            'TEMPLATE_TOP': 'lol',
            'TEMPLATE_BOTTOM': 'otherlol',
            'RANDOM': random.randint(0, 10**18),
        }

    def render_to_string(self, request, template_filename, extra_context={}):
        template = self.env.get_template(template_filename)
        context = {}
        context.update(self.process_context(request))
        context.update(extra_context)
        return template.render(**context)

    def render(self, request, template_filename, extra_context={}):
        response = singletons.server.response
        rendered = self.render_to_string(
            request, template_filename, extra_context)
        return response.html(rendered)



# Modules which are automatically loaded as soon as settings is loaded, to
# ensure singletons etc get hooked up correctly
AUTOLOAD = [
    'omnic.conversion.graph',
    'omnic.cli.commandparser',
    'omnic.cli.commands',
    'omnic.responses.placeholder',
    'omnic.web.server',
    'omnic.web.eventloop',
    'omnic.web.viewer',
    'omnic.worker.manager',
    'omnic.types.detectors',
]

HOST = '127.0.0.1'
PORT = 8080
DEBUG = True

WEB_SERVER = 'sanic'
EVENT_LOOP = 'uvloop'
WORKER = 'omnic.worker.aioworker.AioWorker'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True
        },
    }
}


DETECTORS = [
    'omnic.builtin.types.core.MagicDetector',
    'omnic.builtin.types.core.ExtensionDetector',
    'omnic.builtin.types.core.ManifestDetector',
    'omnic.builtin.types.nodejs.NodePackageDetector',
]

SERVICES = [
    'omnic.builtin.services.media',
    'omnic.builtin.services.test',
    'omnic.builtin.services.admin',
    'omnic.builtin.services.viewer',
]

CONVERTERS = [
    # 'omnic.builtin.converters.thumb.PILThumb', # Disabling Pillow thumb
    'omnic.builtin.converters.thumb.ImageMagickThumb',
    'omnic.builtin.converters.document.Unoconv',
    'omnic.builtin.converters.document.ImageMagickPageRasterizer',
    'omnic.builtin.converters.mesh.MeshLabConverter',
    'omnic.builtin.converters.mesh.Jsc3dRenderer',
    'omnic.builtin.converters.chemical.OpenBabelConverter',
    'omnic.builtin.converters.vector.InkscapeConverter',
    'omnic.builtin.converters.vector.InkscapeRasterizer',
    'omnic.builtin.converters.manifest.ManifestDownloader',

    'omnic.builtin.converters.nodejs.NodePackageDetector',
    'omnic.builtin.converters.nodejs.NPMInstalledDirectoryConverter',
    'omnic.builtin.converters.nodejs.BrowserifyBundler',
    'omnic.builtin.converters.nodejs.BabelES6Compiler',
    'omnic.builtin.converters.nodejs.UglifyJSMinifier',

    'omnic.builtin.services.viewer.converters.ViewerNodePackageBuilder',
]

PREFERRED_CONVERSION_PATHS = []
CONVERSION_PROFILES = {}

VIEWERS = [
    'omnic.builtin.viewers.omnic_viewer_core',
    'omnic.builtin.viewers.simple_image_viewer',
    'omnic.builtin.viewers.simple_pdf_embed',
]

PLACEHOLDERS = [
    'omnic.builtin.responses.placeholders.PNGPixel',
    'omnic.builtin.responses.placeholders.JSGlobalVar',
    'omnic.builtin.responses.placeholders.EmptyAll',
]

PATH_PREFIX = '/tmp/omnic/'
RESOURCE_CACHE_INTERFIX = 'resource'
PATH_GROUPING = 'MD5'
ALLOWED_LOCATIONS = {
    # local
    'localhost', '127.0.0.1',

    # test sites mentioned in README
    'unsplash.it', 'people.sc.fsu.edu', 'imr.sandia.gov', 'wiki.jmol.org',
    'opendatakit.org',
}

# Conversion type safety stuff:
CONVERSION_SYSTEM_CHECK = True



from omnic.conversion import converter


class NodeSdfToSvgRenderer(converter.ExecConverter):
    inputs = [
        'SD',
        'SDF',
        'chemical/x-mdl-sdfile',
    ]

    outputs = [
        'SVG',
        'image/svg+xml',
    ]

    command = [
        'sdftosvg',
        '$IN',
        '$OUT',
    ]


class OpenBabelConverter(converter.ExecConverter):
    # NOTE: All these input types are untested
    inputs = [
        'ABINIT',  # ABINIT Output Format
        'ACESOUT',  # ACES output format
        'ACR',  # ACR format
        'ADFOUT',  # ADF output format
        'ALC',  # Alchemy format
        'ARC',  # Accelrys/MSI Biosym/Insight II CAR format
        'AXSF',  # XCrySDen Structure Format
        'BGF',  # MSI BGF format
        'BOX',  # Dock 3.5 Box format
        'BS',  # Ball and Stick format
        'C09OUT',  # Crystal 09 output format
        'C3D1',  # Chem3D Cartesian 1 format
        'C3D2',  # Chem3D Cartesian 2 format
        'CACCRT',  # Cacao Cartesian format
        'CAN',  # Canonical SMILES format
        'CAR',  # Accelrys/MSI Biosym/Insight II CAR format
        'CASTEP',  # CASTEP format
        'CCC',  # CCC format
        'CDX',  # ChemDraw binary format
        'CDXML',  # ChemDraw CDXML format
        'CIF',  # Crystallographic Information File
        'CK',  # ChemKin format
        'CML',  # Chemical Markup Language
        'CMLR',  # CML Reaction format
        'config',  # DL-POLY CONFIG
        'contcar',  # VASP format
        'CRK2D',  # Chemical Resource Kit diagram(2D)
        'CRK3D',  # Chemical Resource Kit 3D format
        'CT',  # ChemDraw Connection Table format
        'CUB',  # Gaussian cube format
        'CUBE',  # Gaussian cube format
        'DAT',  # Generic Output file format
        'DMOL',  # DMol3 coordinates format
        'DX',  # OpenDX cube format for APBS
        'ENT',  # Protein Data Bank format
        'FA',  # FASTA format
        'FASTA',  # FASTA format
        'FCH',  # Gaussian formatted checkpoint file format
        'FCHK',  # Gaussian formatted checkpoint file format
        'FCK',  # Gaussian formatted checkpoint file format
        'FEAT',  # Feature format
        'FHIAIMS',  # FHIaims XYZ format
        'FRACT',  # Free Form Fractional format
        'FS',  # Fastsearch format
        'FSA',  # FASTA format
        'G03',  # Gaussian Output
        'G09',  # Gaussian Output
        'G92',  # Gaussian Output
        'G94',  # Gaussian Output
        'G98',  # Gaussian Output
        'GAL',  # Gaussian Output
        'GAM',  # GAMESS Output
        'GAMESS',  # GAMESS Output
        'GAMIN',  # GAMESS Input
        'GAMOUT',  # GAMESS Output
        'GOT',  # GULP format
        'GPR',  # Ghemical format
        'GRO',  # GRO format
        'GUKIN',  # GAMESS-UK Input
        'GUKOUT',  # GAMESS-UK Output
        'GZMAT',  # Gaussian Z-Matrix Input
        'HIN',  # HyperChem HIN format
        'history',  # DL-POLY HISTORY
        'INCHI',  # InChI format
        'INP',  # GAMESS Input
        'INS',  # ShelX format
        'JOUT',  # Jaguar output format
        'LOG',  # Generic Output file format
        'MCDL',  # MCDL format
        'MCIF',  # Macromolecular Crystallographic Info
        'MDL',  # MDL MOL format
        'ML2',  # Sybyl Mol2 format
        'MMCIF',  # Macromolecular Crystallographic Info
        'MMD',  # MacroModel format
        'MMOD',  # MacroModel format
        'MOL',  # MDL MOL format
        'MOL2',  # Sybyl Mol2 format
        'MOLD',  # Molden format
        'MOLDEN',  # Molden format
        'MOLF',  # Molden format
        'MOO',  # MOPAC Output format
        'MOP',  # MOPAC Cartesian format
        'MOPCRT',  # MOPAC Cartesian format
        'MOPIN',  # MOPAC Internal
        'MOPOUT',  # MOPAC Output format
        'MPC',  # MOPAC Cartesian format
        'MPO',  # Molpro output format
        'MPQC',  # MPQC output format
        'MRV',  # Chemical Markup Language
        'MSI',  # Accelrys/MSI Cerius II MSI format
        'NWO',  # NWChem output format
        'OUT',  # Generic Output file format
        'OUTMOL',  # DMol3 coordinates format
        'OUTPUT',  # Generic Output file format
        'PC',  # PubChem format
        'PCM',  # PCModel Format
        'PDB',  # Protein Data Bank format
        'PDBQT',  # AutoDock PDQBT format
        'POS',  # POS cartesian coordinates format
        'poscar',  # VASP format
        'PQR',  # PQR format
        'PQS',  # Parallel Quantum Solutions format
        'PREP',  # Amber Prep format
        'PWSCF',  # PWscf format
        'QCOUT',  # Q-Chem output format
        'RES',  # ShelX format
        'RSMI',  # Reaction SMILES format
        'RXN',  # MDL RXN format
        'SD',  # MDL MOL format
        'SDF',  # MDL MOL format
        'SMI',  # SMILES format
        'SMILES',  # SMILES format
        'SY2',  # Sybyl Mol2 format
        'T41',  # ADF TAPE41 format
        'TDD',  # Thermo format
        'TEXT',  # Read and write raw text
        'THERM',  # Thermo format
        'TMOL',  # TurboMole Coordinate format
        # 'TXT', # Title format
        'TXYZ',  # Tinker XYZ format
        'UNIXYZ',  # UniChem XYZ format
        'vasp',  # VASP format
        'VMOL',  # ViewMol format
        'XML',  # General XML format
        'XSF',  # XCrySDen Structure Format
        'XTC',  # XTC format
        'XYZ',  # XYZ cartesian coordinates format
        'YOB',  # YASARA.org YOB format

        # mimetypes of the above
        'chemical/x-alchemy',
        'chemical/x-cdx',
        'chemical/x-cerius',
        'chemical/x-chem3d',
        'chemical/x-chemdraw',
        'chemical/x-cif',
        'chemical/x-cml',
        'chemical/x-daylight-smiles',
        'chemical/x-gamess-input',
        'chemical/x-gaussian-checkpoint',
        'chemical/x-gaussian-cube',
        'chemical/x-gaussian-input',
        'chemical/x-gaussian-log',
        'chemical/x-hin',
        'chemical/x-jcamp-dx',
        'chemical/x-kinemage',
        'chemical/x-macmolecule',
        'chemical/x-macromodel-input',
        'chemical/x-mdl-molfile',
        'chemical/x-mdl-rdfile',
        'chemical/x-mdl-rxnfile',
        'chemical/x-mdl-sdfile',
        'chemical/x-mdl-tgf',
        'chemical/x-mmcif',
        'chemical/x-mol2',
        'chemical/x-mopac-graph',
        'chemical/x-mopac-input',
        'chemical/x-mopac-out',
        'chemical/x-pdb',
        'chemical/x-rosdal',
        'chemical/x-xyz',
    ]

    outputs = [
        'PNG',
        'SVG',
        'image/png',
        'image/svg+xml',
    ]

    command = [
        'obabel',
        '$IN',
        '$0',
    ]

    def get_arguments(self, out):
        return ['-O%s' % out.cache_path]



import argparse
import sys
import textwrap

import omnic
from omnic import singletons
from omnic.cli import consts
from omnic.utils.asynctools import coerce_to_synchronous


class CommandParser:
    def __init__(self):
        self.subcommands = {}
        self._last_args = None

    def gen_subcommand_help(self):
        '''
        Generates s
        '''
        commands = sorted(self.subcommands.items(), key=lambda i: i[0])
        return '\n'.join(
            '%s %s' % (
                subcommand.ljust(15),
                textwrap.shorten(description, width=61),
            )
            for subcommand, (description, action, opts) in commands
        )

    def parse_args(self, argv=None):
        parser = argparse.ArgumentParser(
            prog='omnic',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description='Generalized file conversion framework, useful as '
            'web micro-service.',
            epilog=(
                consts.CLI_HELP_EPILOG_TEMPLATE % self.gen_subcommand_help()
            ),
        )

        parser.add_argument('-V', '--version',
                            action='version',
                            version='%(prog)s v' + omnic.__version__,
                            )

        parser.add_argument('-v', '--verbose',
                            action='store_true',
                            default=False,
                            help='increase output verbosity',
                            )

        subparsers = parser.add_subparsers(dest='subcommand')
        for name, (description, action, opts) in self.subcommands.items():
            subparser = subparsers.add_parser(name, description=description)
            for args, kwargs in opts:
                subparser.add_argument(*args, **kwargs)

        args = parser.parse_args(argv)
        self._last_args = args
        if not args.subcommand:
            parser.print_help()
            sys.exit(1)
        return args

    def parse_args_to_action_args(self, argv=None):
        '''
        Parses args and returns an action and the args that were parsed
        '''
        args = self.parse_args(argv)
        action = self.subcommands[args.subcommand][1]
        return action, args

    def register_subparser(self, action, name, description='', arguments={}):
        '''
        Registers a new subcommand with a given function action.

        If the function action is synchronous
        '''
        action = coerce_to_synchronous(action)
        opts = []
        for flags, kwargs in arguments.items():
            if isinstance(flags, str):
                flags = tuple([flags])
            opts.append((flags, kwargs))
        self.subcommands[name] = (description, action, opts)

    def subcommand(self, description='', arguments={}):
        '''
        Decorator for quickly adding subcommands to the omnic CLI
        '''
        def decorator(func):
            self.register_subparser(
                func,
                func.__name__.replace('_', '-'),
                description=description,
                arguments=arguments,
            )
            return func
        return decorator

    def print(self, *args, **kwargs):
        '''
        Utility function that behaves identically to 'print' except it only
        prints if verbose
        '''
        if self._last_args and self._last_args.verbose:
            print(*args, **kwargs)

    def printerr(self, *args, **kwargs):
        kwargs['file'] = sys.stderr
        print(*args, **kwargs)


singletons.register('cli', CommandParser)



'''
Tests for `resource` module.
'''
import hashlib
import os
import tempfile
from unittest.mock import call, patch

import pytest
import requests_mock

from omnic import singletons
from omnic.config.exceptions import ConfigurationError
from omnic.config.utils import use_settings
from omnic.types.resource import (CacheError, ForeignBytesResource,
                                  ForeignResource, TypedResource, URLError)
from omnic.types.typestring import TypeString

from .testing_utils import Magic, rm_tmp_files

URL = 'http://mocksite.local/file.png'
URL1 = 'http://mocksite.local/otherthing/file.png'
URL2 = 'http://mocksite2.local/otherthing/file.png'
URL_GIF = 'http://mocksite.local/file.png.gif'


class MockConfig:
    PATH_GROUPING = 'MD5'
    PATH_PREFIX = ''
    ALLOWED_LOCATIONS = '*'
    LOGGING = None


class MockConfigRestrictive(MockConfig):
    ALLOWED_LOCATIONS = {'mocksite.local'}


class TestForeignResource:
    def setup_method(self, method):
        MockConfig.PATH_PREFIX = tempfile.mkdtemp(prefix='omnic_test_')
        singletons.settings.use_settings(MockConfig)
        self.res = ForeignResource(URL)

    def teardown_method(self, method):
        singletons.settings.use_previous_settings()
        try:
            os.remove(self.res.cache_path)
        except OSError:
            pass
        try:
            os.removedirs(os.path.dirname(self.res.cache_path))
        except OSError:
            pass

    def test_properties(self):
        assert self.res.url_string == URL
        assert self.res.url_path_basename == 'file.png'
        assert self.res.cache_path.endswith('file.png')
        assert not self.res.cache_exists()

    def test_uniqueness(self):
        res1 = ForeignResource(URL1)
        res2 = ForeignResource(URL2)
        assert self.res.cache_path != res1.cache_path
        assert self.res.cache_path != res2.cache_path
        assert res2.cache_path != res1.cache_path

    def test_download(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', URL, content=Magic.JPEG)
            self.res.download()
            assert self.res.cache_exists()
        with self.res.cache_open() as f:
            assert f.read() == Magic.JPEG

    def test_cache_open(self):
        with self.res.cache_open('wb') as f:
            f.write(b'test')
        with self.res.cache_open() as f:
            assert f.read() == b'test'

    def test_type_identification_exception(self):
        with pytest.raises(CacheError):
            self.res.guess_typed()

    def test_type_identification(self):
        with requests_mock.mock() as m:
            m.register_uri('GET', URL, content=Magic.JPEG)
            self.res.download()
            assert self.res.cache_exists()
        tr = self.res.guess_typed()
        assert tr.typestring.mimetype == 'image/jpeg'


class TestForeignResourceDirectory:
    def setup_method(self, method):
        MockConfig.PATH_PREFIX = tempfile.mkdtemp(prefix='omnic_test_')
        singletons.settings.use_settings(MockConfig)
        self.res = ForeignResource(URL)

    def teardown_method(self, method):
        singletons.settings.use_previous_settings()
        try:
            os.remove(os.path.join(self.res.cache_path, 'thing', 'testfile'))
        except OSError:
            pass
        os.removedirs(os.path.join(self.res.cache_path, 'thing'))

    def test_cache_open_as_dir(self):
        with self.res.cache_open_as_dir('thing/testfile', 'wb') as f:
            f.write(b'test')
        with self.res.cache_open_as_dir('thing/testfile') as f:
            assert f.read() == b'test'


class TestResourceValidate:
    def test_validate_global(self):
        res = ForeignResource(URL)
        res1 = ForeignResource(URL1)
        res2 = ForeignResource(URL2)
        singletons.settings.use_settings(MockConfig)
        res.validate()  # no exceptions since config is loose
        res1.validate()  # no exceptions since config is loose
        res2.validate()  # no exceptions since config is loose

    def test_validate_restrictive(self):
        singletons.settings.use_settings(MockConfigRestrictive)
        res = ForeignResource(URL)
        res1 = ForeignResource(URL1)
        res2 = ForeignResource(URL2)
        res.validate()  # no exceptions since okay
        res1.validate()  # no exceptions since okay
        with pytest.raises(URLError):
            res2.validate()  # exception since different domain


class TestTypedResource:
    @classmethod
    def setup_class(cls):
        singletons.settings.use_settings(MockConfig)
        cls.res = TypedResource(URL, TypeString('image/gif'))

    @classmethod
    def teardown_class(cls):
        singletons.settings.use_previous_settings()

    def test_properties(self):
        assert self.res.url_string == URL
        assert self.res.cache_path.endswith('file.png.gif')
        assert not self.res.cache_exists()

    def test_uniqueness(self):
        paths = set([
            TypedResource(URL, TypeString(
                'image/gif')).cache_path,
            TypedResource(URL, TypeString(
                'image/png')).cache_path,
            TypedResource(URL_GIF,
                          TypeString('image/gif')).cache_path,
            TypedResource(URL1, TypeString(
                'image/gif')).cache_path,
            TypedResource(URL2, TypeString(
                'image/gif')).cache_path,
        ])
        assert len(paths) == 5


class TestForeignBytesResource:
    def setup_method(self, method):
        MockConfig.PATH_PREFIX = tempfile.mkdtemp(prefix='omnic_test_')
        singletons.settings.use_settings(MockConfig)

    def teardown_method(self, method):
        singletons.settings.use_previous_settings()
        rm_tmp_files(self.res.cache_path)

    def _check_md5(self, res, data):
        md5 = hashlib.md5(data).hexdigest()
        assert not res.cache_exists()
        assert md5 in res.url_string
        assert res.url_string.startswith('file://')
        with pytest.raises(CacheError):
            res.guess_typed()

    def test_without_any_hints(self):
        data = b'      '  # strange data
        res = ForeignBytesResource(data)
        self.res = res
        self._check_md5(res, data)
        assert 'source' in res.url_string
        self.res.save()
        typed = res.guess_typed()
        assert str(typed.typestring) == 'unknown'

    def test_basename(self):
        self.res = ForeignBytesResource(b'', basename='input')
        assert 'input' in self.res.url_string

    def test_extension(self):
        data = b'      '  # strange data
        self.res = ForeignBytesResource(data, extension='txt')
        self._check_md5(self.res, data)
        self.res.download()
        typed = self.res.guess_typed()
        assert str(typed.typestring) == 'TXT'

    def test_guess_type(self):
        data = Magic.JPEG
        self.res = ForeignBytesResource(data)
        self._check_md5(self.res, data)
        assert 'source' in self.res.url_string
        self.res.download()
        typed = self.res.guess_typed()
        assert str(typed.typestring) == 'image/jpeg'


class TestClearCache:
    def test_remove_cache(self):
        res = TypedResource(URL, TypeString('image/gif'))
        with patch('os.unlink') as mock:
            res.cache_remove()
        assert mock.mock_calls == [call(res.cache_path)]

    def test_remove_dircache(self):
        res = TypedResource(URL, TypeString('directory'))
        with patch('shutil.rmtree') as mock:
            res.cache_remove_as_dir()
        assert mock.mock_calls == [call(res.cache_path)]

    def test_remove_dircache(self):
        res = ForeignResource(URL)
        with patch('shutil.rmtree') as mock:
            res.cache_remove_all()
        assert mock.mock_calls == [call(os.path.dirname(res.cache_path))]


class TestCacheSettings:
    @use_settings(path_grouping='INVALID')
    def test_invalid_path_grouping(self):
        with pytest.raises(ConfigurationError):
            ForeignResource(URL)



import json
import os

from omnic.builtin.types.core import DIRECTORY
from omnic.builtin.types.nodejs import (INSTALLED_NODE_PACKAGE,
                                        INSTALLED_WEBPACK_NODE_PACKAGE,
                                        NODE_PACKAGE, WEBPACK_NODE_PACKAGE,
                                        NodePackageDetector)
from omnic.conversion import converter


class NodePackageDetector(converter.DetectorConverter):
    detector = NodePackageDetector
    inputs = [
        str(DIRECTORY),
    ]
    outputs = [
        str(NODE_PACKAGE),
        str(WEBPACK_NODE_PACKAGE),
        str(INSTALLED_NODE_PACKAGE),
        str(INSTALLED_WEBPACK_NODE_PACKAGE),
    ]


class NPMInstalledDirectoryConverter(converter.AdditiveDirectoryExecConverter):
    inputs = [
        str(NODE_PACKAGE),
        str(WEBPACK_NODE_PACKAGE),
    ]
    outputs = [
        str(INSTALLED_NODE_PACKAGE),
        str(INSTALLED_WEBPACK_NODE_PACKAGE),
    ]
    command = [
        'npm',
        'install',
    ]


class BrowserifyBundler(converter.ExecConverter):
    inputs = [
        str(INSTALLED_NODE_PACKAGE),
    ]
    outputs = [
        'bundle.js',
    ]

    def get_command(self, in_resource, out_resource):
        # Get probable name of main JS entry point
        pjson_path = os.path.join(in_resource.cache_path, 'package.json')
        package_json = json.load(open(pjson_path))
        main_js = package_json.get('main', 'index.js')
        main_js_path = os.path.join(in_resource.cache_path, main_js)
        return [
            'browserify',
            main_js_path,
            '-o',
            out_resource.cache_path,
        ]


class BabelES6Compiler(converter.ExecConverter):
    # NOTE: npm install -g babel babel-cli babel-preset-es2015
    # NOTE: It's a terrible hack to use es2015 globally, but it is the omnic
    # way, and will pay off once I build out precise containers
    inputs = [
        'JS',
        'bundle.js',
    ]
    outputs = [
        'dev-es5.js',
        'es5.js',
    ]

    def get_command(self, in_resource, out_resource):
        extra_args = []
        out = str(out_resource.typestring)

        if 'dev' in out:
            extra_args.extend(['--source-maps', 'inline'])

        if 'es5' in out:
            # TODO XXX obviously broken hack: Must fix by adding the
            # "configure" step mechanism, that gets the absolute path to
            # whichever programs/libs/whatever are necessary, and also what is
            # supported
            es2015_path = os.path.expanduser('~/.local/lib/node_modules'
                                             '/babel-preset-es2015')
            extra_args.extend(['--presets=%s' % es2015_path])

        return [
            'babel',
            '--no-babelrc',
            in_resource.cache_path,
            '--out-file',
            out_resource.cache_path,
            *extra_args,
        ]


class UglifyJSMinifier(converter.ExecConverter):
    # NOTE: npm install -g uglify-js
    inputs = [
        'es5.js',
    ]
    outputs = [
        'min.js',
    ]

    command = [
        'uglifyjs',
        '$IN',
        '-o',
        '$OUT',
    ]



import json
import os

from omnic import singletons
from omnic.types.resource import ForeignBytesResource

VIEWER_EXT = 'omnic_viewer_descriptor'  # TODO Hack


class ViewerManager:
    def __init__(self, viewer_list=None):
        settings = singletons.settings
        self.viewers = viewer_list
        if self.viewers is None:
            self.viewers = settings.load_all('VIEWERS', default=DefaultViewer)

    def prefix_asset(self, viewer, relpath):
        return os.path.join(viewer.asset_dir, relpath)

    def get_assets(self):
        '''
        Return a flat list of absolute paths to all assets required by this
        viewer
        '''
        return sum([
            [self.prefix_asset(viewer, relpath) for relpath in viewer.assets]
            for viewer in self.viewers
        ], [])

    def get_node_packages(self):
        '''
        Return a flat list of absolute paths to all node modules required by
        this viewer
        '''
        return {
            viewer.name: viewer.node_package
            for viewer in self.viewers
        }

    def get_resource(self):
        '''
        Returns a BytesResource to build the viewers JavaScript
        '''
        # Basename could be used for controlling caching
        # basename = 'viewers_%s' % settings.get_cache_string()
        node_packages = self.get_node_packages()
        # sort_keys is essential to ensure resulting string is
        # deterministic (and thus hashable)
        viewers_data_str = json.dumps(node_packages, sort_keys=True)
        viewers_data = viewers_data_str.encode('utf8')
        viewers_resource = ForeignBytesResource(
            viewers_data,
            extension=VIEWER_EXT,
            # basename=basename,
        )
        return viewers_resource


class DefaultViewer:
    def __init__(self, import_path):
        _package_dir = singletons.settings \
            .import_path_to_absolute_path(import_path, 'package.json')
        _package_json_path = os.path.join(_package_dir, 'package.json')

        with open(_package_json_path) as fd:
            package_json = json.load(fd)

        self.types = package_json['omnic']['types']
        self.name = package_json['name']
        self.asset_dir = _package_dir
        self.assets = []
        self.node_package = 'file:%s' % _package_dir


singletons.register('viewers', ViewerManager)



from itertools import zip_longest

START = object()  # just using as a unique symbol


def pair_looper(iterator):
    '''
    Loop through iterator yielding items in adjacent pairs
    '''
    left = START
    for item in iterator:
        if left is not START:
            yield (left, item)
        left = item


def first_last_iterator(iterable):
    lst = list(iterable)
    length = len(lst)
    for index, item in enumerate(lst):
        yield index == 0, index == (length - 1), item


def group_by(iterable, n, fill=None):
    args = [iter(iterable)] * n
    return list(''.join(s) for s in zip_longest(*args, fillvalue=fill))



import logging

import aiohttp

from omnic import singletons
from omnic.worker import tasks
from omnic.worker.enums import Task

log = logging.getLogger()

DOWNLOAD_TIMEOUT = 20
DOWNLOAD_CHUNK_SIZE = 8124


class BaseWorker:
    '''
    Worker base class
    '''

    def __init__(self):
        self.stats_dequeued = 0
        self.stats_began = 0
        self.stats_success = 0
        self.stats_error = 0

    def __del__(self):
        self._close()

    def _close(self):
        if hasattr(self, 'aiohttp'):
            if not self.aiohttp.closed:
                self.aiohttp.close()

    async def run(self):
        while self.running:
            # Queue up consuming next item
            task_type, args = await self.get_next()
            self.stats_dequeued += 1
            method = self._get_method(task_type)

            # Determine the type of task, and possibly skip if we have
            # it "locked" that we are already doing it
            if task_type == Task.FUNC:
                pass

            elif task_type == Task.DOWNLOAD:
                if not await self.check_download(*args):
                    log.debug('Already downloading %s' % repr(args))
                    continue

            elif task_type == Task.CONVERT:
                if not await self.check_convert(*args):
                    log.debug('Already converting %s' % repr(args))
                    continue

            elif task_type == Task.MULTICONVERT:
                if not await self.check_multiconvert(*args):
                    log.debug('Already multiconverting %s' % repr(args))
                    continue

            # Queue it up and run it
            self.stats_began += 1
            try:
                await method(*args)
                self.stats_success += 1
            except Exception as e:
                self.stats_error += 1
                log.exception('Error in task: "%s"' % repr(e))

    def _get_method(self, task_type):
        return {
            Task.FUNC: self.run_func,
            Task.DOWNLOAD: self.run_download,
            Task.CONVERT: self.run_convert,
            Task.MULTICONVERT: self.run_multiconvert,
        }.get(task_type)

    async def run_func(self, func, *func_args):
        '''
        Runs arbitrary synchronous code
        '''
        func(*func_args)

    async def run_download(self, foreign_resource):
        '''
        Downloads a foreign resource asynchronously
        '''
        url = foreign_resource.url_string
        with foreign_resource.cache_open('wb') as f_handle:
            await self._download_async(url, f_handle)

    async def _download_async(self, url, f_handle):
        loop = singletons.eventloop.loop
        self.aiohttp = aiohttp.ClientSession(loop=loop)
        with aiohttp.Timeout(DOWNLOAD_TIMEOUT):
            async with self.aiohttp.get(url) as response:
                while True:
                    chunk = await response.content.read(DOWNLOAD_CHUNK_SIZE)
                    if not chunk:
                        break
                    f_handle.write(chunk)
                self._close()
                return await response.release()

    async def run_convert(self, converter, in_resource, out_resource):
        '''
        Converts using the given converter, asynchronously if available,
        otherwise falls back on sync
        '''
        if hasattr(converter, 'convert'):
            await converter.convert(in_resource, out_resource)
        elif hasattr(converter, 'convert_sync'):
            converter.convert_sync(in_resource, out_resource)
        else:
            raise ValueError('Invalid converter: %s' % repr(converter))

    async def run_multiconvert(self, url_string, to_type):
        '''
        Enqueues in succession all conversions steps necessary to take the
        given URL and convert it to to_type, storing the result in the cache
        '''
        async def enq_convert(*args):
            await self.enqueue(Task.CONVERT, args)
        await tasks.multiconvert(url_string, to_type, enq_convert)



import os

from omnic.conversion import converter


class Unoconv(converter.ExecConverter):
    inputs = [
        # Document
        'BIB',
        'DOC',
        'HTML',
        'XHTML',
        'HTM',
        'ODT',
        'OTT',
        'RTF',
        'LTX',
        'SDW',
        'STW',
        'SXW',
        'TXT',
        'UOT',
        'VOR',

        # Spreadsheet
        'CSV',
        'XLS',
        'DBF',
        'ODS',
        'XML',

        # Presentation
        'ODP',
        'PPT',
        'PPTX',

        'application/msword',
        'application/vnd.oasis.opendocument.text',
    ]

    outputs = [
        'PDF',
        'application/pdf',
    ]

    command = [
        # TODO Fix this, has to do with virtualenv issues
        '/usr/bin/python3',
        '/usr/bin/unoconv',
        '-f',
        '$0',
        '-o',
        '$OUT',
        '$IN',
    ]

    def get_arguments(self, out_resource):
        return [out_resource.typestring.extension.lower()]


class ImageMagickPageRasterizer(converter.ExecConverter):
    inputs = [
        # Document
        'PDF',
        'PS',
        'application/pdf',
        'application/postscript',
    ]

    outputs = [
        'PNG',
        'image/png',
    ]

    command = [
        'convert',
        '$IN',
        '$0',
    ]

    def get_arguments(self, out_resource):
        page_number = 0
        source = '%s[%i]' % (out_resource.cache_path, page_number)
        return [source]

    def get_output_filename(self, in_resource, out_resource):
        dirname = os.path.dirname(out_resource.cache_path)
        basename = os.path.basename(out_resource.cache_path)
        basename_base, _ = os.path.splitext(basename)
        suffix = '0.png'
        return os.path.join(dirname, '%s-%s' % (basename_base, suffix))
